import pytest
from gym.spaces import Tuple, Dict, Box, Discrete
from godot_rl.core.godot_env import GodotEnv
from godot_rl.core.utils import ActionSpaceProcessor

@pytest.mark.parametrize("action_space", 
    [
        Tuple([Box(-1,1, shape=[7]), Box(-1,1, shape=[11])]),
        Tuple([Box(-1,1, shape=[7]), Discrete(2)]),
        Tuple([Discrete(2), Discrete(2)]),
        Tuple([Discrete(2), Discrete(2), Box(-1,1, shape=[11])]),
    ]



)
def test_action_space_preprocessor(action_space):

    expected_output = 0

    for space in action_space.spaces:
        if isinstance(space, Box):
            assert len(space.shape) ==1
            expected_output += space.shape[0]
        elif isinstance(space, Discrete):
            if space.n > 2:
                # for not only binary actions are supported, need to add support for the n>2 case
                raise NotImplementedError
            expected_output += 1

    preprocessor = ActionSpaceProcessor(action_space, True)

    assert expected_output == preprocessor.action_space.shape[0]

    preprocessor = ActionSpaceProcessor(action_space, False)
    for s1, s2 in zip(action_space.spaces, preprocessor.action_space.spaces):
        if isinstance(s1, Box):
            assert s1.shape == s2.shape
        elif isinstance(s2, Discrete):
            assert s1.n == s2.n



import pytest

from godot_rl.core.godot_env import GodotEnv


@pytest.mark.parametrize(
    "env_name,port,n_agents",
    [
        (
            "BallChase",
            12008,
            16,
        ),
        (
            "FPS",
            12009,
            8,
        ),
        (
            "JumperHard",
            12010,
            16,
        ),
        (
            "Racer",
            12011,
            8,
        ),
        (
            "FlyBy",
            12012,
            16,
        ),
    ],
)
def test_env_ij(env_name, port, n_agents):
    env_path = f"examples/godot_rl_{env_name}/bin/{env_name}.x86_64"
    env = GodotEnv(env_path=env_path, port=port)

    action_space = env.action_space
    observation_space = env.observation_space
    n_envs = env.num_envs

    for j in range(2):
        obs, info = env.reset()
        assert len(obs) == n_envs
        for i in range(10):
            action = [action_space.sample() for _ in range(n_envs)]
            obs, reward, term, trunc, info = env.step(action, order_ij=True)

            assert len(obs) == n_envs
            assert len(reward) == n_envs
            assert len(term) == n_envs
            assert len(trunc) == n_envs
            assert len(info) == n_envs

            assert isinstance(
                reward[0], (float, int)
            ), f"The reward returned by 'step()' must be a float or int, and is {reward[0]} of type {type(reward[0])}"
            assert isinstance(term[0], bool), f"The 'done' signal {term[0]}  {type(term[0])} must be a boolean"
            assert isinstance(info[0], dict), "The 'info' returned by 'step()' must be a python dictionary"

    env.close()


@pytest.mark.parametrize(
    "env_name,port,n_agents",
    [
        (
            "BallChase",
            13008,
            16,
        ),
        (
            "FPS",
            13009,
            8,
        ),
        (
            "JumperHard",
            13010,
            16,
        ),
        (
            "Racer",
            13011,
            8,
        ),
        (
            "FlyBy",
            13012,
            16,
        ),
    ],
)
def test_env_ji(env_name, port, n_agents):
    env_path = f"examples/godot_rl_{env_name}/bin/{env_name}.x86_64"
    env = GodotEnv(env_path=env_path, port=port)

    action_space = env.action_space
    observation_space = env.observation_space
    n_envs = env.num_envs
    assert n_envs == n_agents
    for j in range(2):
        obs, info = env.reset()
        assert len(obs) == n_envs
        for i in range(10):
            action = [action_space.sample() for _ in range(n_envs)]
            action = list(zip(*action))
            obs, reward, term, trunc, info = env.step(action)

            assert len(obs) == n_envs
            assert len(reward) == n_envs
            assert len(term) == n_envs
            assert len(trunc) == n_envs
            assert len(info) == n_envs

            assert isinstance(
                reward[0], (float, int)
            ), f"The reward returned by 'step()' must be a float or int, and is {reward[0]} of type {type(reward[0])}"
            assert isinstance(term[0], bool), f"The 'done' signal {term[0]}  {type(term[0])} must be a boolean"
            assert isinstance(info[0], dict), "The 'info' returned by 'step()' must be a python dictionary"

    env.close()




import numpy as np
import gym
from godot_rl.core.utils import lod_to_dol
from godot_rl.core.godot_env import GodotEnv


class CleanRLGodotEnv:
    def __init__(self, env_path=None, convert_action_space=False, **kwargs):
        # convert_action_space: combine multiple continue action spaces into one larger space 
        self._env = GodotEnv(env_path=env_path,convert_action_space=convert_action_space, **kwargs)
        

    def _check_valid_action_space(self):
        action_space = self._env.action_space
        if isinstance(action_space, gym.spaces.Tuple):
            assert (
                len(action_space.spaces) == 1
            ), f"clearn rl supports a single action space, this env constains multiple spaces {action_space}"

    @staticmethod
    def action_preprocessor(action):
        return action

    def step(self, action):
        action = self.action_preprocessor(action)
        obs, reward, term, trunc, info = self._env.step(action)
        obs = lod_to_dol(obs)
        return np.stack(obs["obs"]), reward, term, trunc, info

    def reset(self, seed):
        obs, info = self._env.reset(seed)
        obs = lod_to_dol(obs)
        return np.stack(obs["obs"]), info

    @property
    def single_observation_space(self):
        return self._env.observation_space["obs"]

    @property
    def single_action_space(self):
        return self._env.action_space
        

    @property
    def num_envs(self):
        return self._env.num_envs


import pytest

from godot_rl.core.godot_env import GodotEnv
from godot_rl.main import get_args

try:
    from godot_rl.wrappers.stable_baselines_wrapper import stable_baselines_training
except ImportError as e:

    def stable_baselines_training(args, extras):
        print("Import error when trying to use sb3, this is probably not installed try pip install godot-rl[sb3]")


def test_sb3_training():
    args, extras = get_args()
    args.env = "gdrl"
    args.env_path = "examples/godot_rl_JumperHard/bin/JumperHard.x86_64"

    stable_baselines_training(args, extras, n_steps=10000)



# we download examples from github and we save them in the examples folder

import os
import shutil
from sys import platform
import wget
from zipfile import ZipFile

BANCHES = {"4" : "main",
           "3" : "godot3.5"}

BASE_URL="https://github.com/edbeeching/godot_rl_agents_examples"

def download_examples():
    #select branch
    print("Select Godot version:")
    for key in BANCHES.keys():
        print(f"{key} : {BANCHES[key]}")
    
    branch = input("Enter your choice: ")
    BRANCH = BANCHES[branch]  
    os.makedirs("examples", exist_ok=True)
    URL=f"{BASE_URL}/archive/refs/heads/{BRANCH}.zip"
    print(f"downloading examples from {URL}")
    wget.download(URL, out="")
    print()
    print(f"unzipping")
    with ZipFile(f"{BRANCH}.zip", 'r') as zipObj:
    # Extract all the contents of zip file in different directory
        zipObj.extractall('examples/')
    print(f"cleaning up")
    os.remove(f"{BRANCH}.zip")
    print(f"moving files")
    for file in os.listdir(f"examples/godot_rl_agents_examples-{BRANCH}"):
        shutil.move(f"examples/godot_rl_agents_examples-{BRANCH}/{file}", "examples")
    os.rmdir(f"examples/godot_rl_agents_examples-{BRANCH}")
    


import argparse
import sys
from functools import partial
import random
import numpy as np
from gym import spaces
from sample_factory.cfg.arguments import parse_full_cfg, parse_sf_args
from sample_factory.envs.env_utils import register_env
from sample_factory.train import run_rl
from sample_factory.enjoy import enjoy

from godot_rl.core.godot_env import GodotEnv
from godot_rl.core.utils import lod_to_dol


class SampleFactoryEnvWrapperBatched(GodotEnv):
    @property
    def unwrapped(self):
        return self

    @property
    def num_agents(self):
        return self.num_envs

    def reset(self, seed=None):
        obs, info = super().reset(seed=seed)
        obs = lod_to_dol(obs)
        return {k: np.array(v) for k, v in obs.items()}, info

    def step(self, action):
        obs, reward, term, trunc, info = super().step(action, order_ij=False)
        obs = lod_to_dol(obs)
        return {k: np.array(v) for k, v in obs.items()}, np.array(reward), np.array(term), np.array(trunc) * 0, info

    @staticmethod
    def to_numpy(lod):

        for d in lod:
            for k, v in d.items():
                d[k] = np.array(v)

        return lod

    def render():
        return


class SampleFactoryEnvWrapperNonBatched(GodotEnv):
    @property
    def unwrapped(self):
        return self

    @property
    def num_agents(self):
        return self.num_envs

    def reset(self, seed=None):
        obs, info = super().reset(seed=seed)
        return self.to_numpy(obs), info

    def step(self, action):
        obs, reward, term, trunc, info = super().step(action, order_ij=True)

        return self.to_numpy(obs), np.array(reward), np.array(term), np.array(trunc) * 0, info

    @staticmethod
    def to_numpy(lod):

        for d in lod:
            for k, v in d.items():
                d[k] = np.array(v)

        return lod

    def render():
        return


def make_godot_env_func(env_path, full_env_name, cfg=None, env_config=None, render_mode=None, speedup=1, viz=False):
    seed = 0
    port = 21008 + cfg.base_port
    print("BASE PORT ", cfg.base_port)
    show_window = False
    if env_config:
        port += 1 + env_config.env_id
        seed += 1 + env_config.env_id
        print("env id", env_config.env_id)
        if viz:  #
            print("creating viz env")
            show_window = env_config.env_id == 0
    if cfg.batched_sampling:
        env = SampleFactoryEnvWrapperBatched(
            env_path=env_path, port=port, seed=seed, show_window=show_window, speedup=speedup
        )
    else:
        env = SampleFactoryEnvWrapperNonBatched(
            env_path=env_path, port=port, seed=seed, show_window=show_window, speedup=speedup
        )

    return env


def register_gdrl_env(args):
    make_env = partial(make_godot_env_func, args.env_path, speedup=args.speedup, viz=args.viz)
    register_env("gdrl", make_env)


def gdrl_override_defaults(_env, parser):
    """RL params specific to Atari envs."""
    parser.set_defaults(
        # let's set this to True by default so it's consistent with how we report results for other envs
        # (i.e. VizDoom or DMLab). When running evaluations for reports or to compare with other frameworks we can
        # set this to false in command line
        summaries_use_frameskip=True,
        use_record_episode_statistics=True,
        gamma=0.99,
        env_frameskip=1,
        env_framestack=4,
        num_workers=1,
        num_envs_per_worker=2,
        worker_num_splits=2,
        env_agents=16,
        train_for_env_steps=1000000,
        nonlinearity="relu",
        kl_loss_coeff=0.0,
        use_rnn=False,
        adaptive_stddev=True,
        reward_scale=1.0,
        with_vtrace=False,
        recurrence=1,
        batch_size=2048,
        rollout=32,
        max_grad_norm=0.5,
        num_epochs=2,
        num_batches_per_epoch=4,
        ppo_clip_ratio=0.2,
        value_loss_coeff=0.5,
        exploration_loss="entropy",
        exploration_loss_coeff=0.000,
        learning_rate=0.00025,
        lr_schedule="linear_decay",
        shuffle_minibatches=False,
        gae_lambda=0.95,
        batched_sampling=False,
        normalize_input=True,
        normalize_returns=True,
        serial_mode=False,
        async_rl=True,
        experiment_summaries_interval=3,
        adam_eps=1e-5,
    )


def add_gdrl_env_args(_env, p: argparse.ArgumentParser, evaluation=False):
    if evaluation:
        # apparently env.render(mode="human") is not supported anymore and we need to specify the render mode in
        # the env actor
        p.add_argument("--render_mode", default="human", type=str, help="")
    p.add_argument("--base_port", default=0, type=int, help="")

    p.add_argument(
        "--env_agents",
        default=2,
        type=int,
        help="Num agents in each envpool (if used)",
    )


def parse_gdrl_args(argv=None, evaluation=False):
    parser, partial_cfg = parse_sf_args(argv=argv, evaluation=evaluation)
    add_gdrl_env_args(partial_cfg.env, parser, evaluation=evaluation)
    gdrl_override_defaults(partial_cfg.env, parser)
    final_cfg = parse_full_cfg(parser, argv)
    return final_cfg


def sample_factory_training(args, extras):
    register_gdrl_env(args)
    cfg = parse_gdrl_args(argv=extras, evaluation=args.eval)
    cfg.base_port = random.randint(20000, 22000)
    status = run_rl(cfg)
    return status


def sample_factory_enjoy(args, extras):
    register_gdrl_env(args)
    cfg = parse_gdrl_args(argv=extras, evaluation=args.eval)

    status = enjoy(cfg)
    return status



import time

from godot_rl.core.godot_env import GodotEnv

if __name__ == "__main__":
    show_window = False
    N_STEPS = 100
    env_path = "envs/example_envs/builds/BallChase/ball_chase.x86_64"

    results = {}

    env = GodotEnv()

    obs = env.reset()
    obs, reward, done, info = env.step([env.action_space.sample() for _ in range(env.num_envs)])

    returns = env.call("remote_callable")

    print(returns)
